from WikidataTemplater.DBHandler import DBhandler
from WikidataTemplater.WDhandler import WDhandler
from WikidataTemplater.main import Settings
import json, yaml, os

s = Settings()
FALLBACK_LANGUAGE = s.settings["FALLBACK_LANGUAGE"]
HOME_VARIABLE = s.settings["HOME_VARIABLE"]
DATA_MODEL_FILE = s.settings["DATA_MODEL_FILE"]
DATA_MODEL_DIR = s.settings["DATA_MODEL_DIR"]



class DataModel():
    def __init__(self):
        self.paths = []
        self.statements = []
        

    def build_paths(self):
        '''creates a dict which represents the entire data model'''
        entity = "?home"
        model={}

        model = self.get_children_recursive(model, entity, "?shows", 0)
        return model.copy()

    def create_table(self):
        dbh = DBhandler()
        st_drop = "DROP TABLE IF EXISTS data_model" 
        st_create =  '''
            CREATE TABLE data_model (
                subject TEXT NOT NULL,
                predicate TEXT NOT NULL,
                object TEXT NOT NULL,
                object_label TEXT

            );
                '''
        dbh.execute(st_drop)
        dbh.execute(st_create)

    def find_path(self, entity, path):
        '''traces the way from an entity to home
        '''
        r = self.select((None, None, entity))
        if entity == HOME_VARIABLE:
            path += "/" +  HOME_VARIABLE 
            
            path = self.reverse_path(path)
            return path
        else:
            path += "/" + entity 
            path = self.find_path(r[0][0], path)   
            return path
        if len(r) == 0:
            return path 
        
    def get_children(self,entity, predicate = None):
        d = self.select_to_dict((entity, predicate, None))
        return d.copy()

    def get_children_recursive(self,model, entity, predicate = None, stop = None):
        d = self.select_to_dict((entity, predicate, None))
        stop +=1
        for k, v in d.items():
            for kk,vv in v.items():
                if stop < 10:
                    v[kk] = self.get_children_recursive(d.copy(), kk,predicate, stop).copy()
                    pass
        #print(json.dumps(d, indent=4))
        return d.copy()
    
    def get_entity_label(self, entity, language = FALLBACK_LANGUAGE):
        if entity[0] != "?":
            entity = "?" + entity
        r = self.select ((entity, "?label_" + language, None))
        
        if len(r) > 0:
            
            return r[0][2].replace("_"," ")
        else:
            return entity[1:2].upper() + entity[2:].lower().replace("_"," ")
    
   
    
    def import_statements(self):
        self.read_yaml()
        self.create_table()
        self.write_db()

    def is_singular(self, entity):
        r = self.select((None, "?singular", entity))
        if len(r)>0:
            return True
        else:
            return False

    
       
    
    def plural_of(self, entity):
        r = self.select((None, "?singular", entity))
        if len(r)>0:
            return r[0][0]
        else:
            return ""
        
    def Xread_yaml(self):
        with open(DATA_MODEL_FILE) as f:
            self.statements = yaml.load(f, Loader=yaml.FullLoader)["statements"] 

    def read_yaml(self):
        for filename in os.scandir(DATA_MODEL_DIR):
            if filename.is_file():
                with open(filename) as f:
                    for line in f:
                        s = line.rstrip()
                        if len(s) > 0 and s[0] == "?":
                            self.statements.append(s)
                    
                    

    def reverse_path(self, path):
        path = path.split("/")[::-1]
        return "/".join(path)

        
    def select(self, triple ):
        ''' emulates a SPRQL Select statement
        triple = (subject, predicate, object) where empty parts are None
        '''
        dbh = DBhandler()
        q = 'select * from data_model'
        conditions = []
        for i,e in enumerate(triple):
            if e:
                if i == 0:
                    conditions.append ('subject = "' + e +'" ')
                if i == 1:
                    conditions.append ('predicate = "' + e +'" ')
                if i == 2:
                    conditions.append ('object = "' + e +'" ')
        where = " AND ".join(conditions)
        if len(where) > 0 :
            q = q + " WHERE " + where
        r = dbh.get(q, "fetchall")
        return r
    
    def select_to_dict(self, triple):
        r = self.select(triple)
        d = {}
        for e in r:
            if "p:"+e[1] in d:
                d["p:"+e[1]][e[2]] = {}
            else:
                d["p:"+e[1]] = {}
                d["p:"+e[1]][e[2]] = {}

        return d.copy()

    
    def singular_of(self, entity):
        r = self.select((entity, "?singular", None))
        if len(r)>0:
            return r[0][2]
        else:
            return ""

    def write_db(self):
        dbh = DBhandler()
        q = "INSERT INTO data_model (subject, predicate, object, object_label) VALUES (?,?,?,?)"
        for e in self.statements:
            e = e.replace("  ", " ")
            a = e.split(" ")
            subject = a[0]
            predicate = a[1]
            object = a[2]
            if len(a) == 4:
                object_label = a[3]
            else:
                object_label = ""
            dbh.insert(q, (subject, predicate, object, object_label))