from WikidataTemplater.View import View
from WikidataTemplater.Presentation import Presentation
from WikidataTemplater.Route import Route
from WikidataTemplater.WDhandler import WDhandler
from WikidataTemplater.Render import Render
from WikidataTemplater.DBHandler import DBhandler
from WikidataTemplater.ErrorHandler import ErrorHandler
from urllib.parse import quote
import re


class ListView(View):
    def __init__(self, view_instance ):
        vars(self).update(vars(view_instance))
        #super.__init__()
        self.count_items = 0
        self.count_pages = 1
        self.pagination = True
        self.pagination_members= []
        self.page = 1
        self.items_per_page = 30
        self.head = []
        self.sort_by: str| None = None
        #self.order_desc = False
        self.items_on_page = []
        self.item_labels = tuple([x["itemLabel"]["value"] for x in self.response["results"]["bindings"]])
        self.rendered = ""



    def order_by(self) -> None:
        """ sort the items of the Wikidata response by name or by 'age'
            Sorting by age means sorting by the numeric value of Wikidata identifier.
        """        
        a: list = self.sort_by.split("|")
        l: list = self.response["results"]["bindings"]
        if a[0] == "wdid":
            self.response["results"]["bindings"] = sorted(l, key=lambda x: int(x["item"]['value'].split("/")[-1:][0][1:]), reverse = bool(int(a[1])))
        if a[0] == "itemLabel":
            self.response["results"]["bindings"] = sorted(l, key=lambda x: x["itemLabel"]['value'], reverse = bool(int(a[1])))

    def paginate(self, page = None):
        if page:
            self.page = int(page)
        else:
            self.page = 1
        self.count_items = len(self.response["results"]["bindings"])
        
        self.count_pages = int(self.count_items / self.items_per_page) + 1
        
        if self.page > int(self.count_items / self.items_per_page) + 1:
            self.page = int(self.count_items / self.items_per_page) +1
        end = self.page * self.items_per_page 
        start = end - self.items_per_page 
        self.items_on_page = self.response["results"]["bindings"][start:end]
        #self.instances_of_items_on_pages = self.get_instances()
        self.prepare_pagination()
        
    def prepare_pagination(self):

        if self.page == 1 :
            self.pagination_members.append(("previous", "desabled", "-"))
        else: 
            self.pagination_members.append(("previous", "", "-"))
        if  self.page ==1:
            self.pagination_members.append(("1", "active", 1)) 
        else:
            self.pagination_members.append(("1", "", 1)) 
        if self.page > 4:
            self.pagination_members.append(("…", "", self.page)) 
        start = self.page -3
        end = self.page + 3
        if start < 2:
            start = 2
        if end > self.count_pages - 1:
            end = self.count_pages -1
        for i in range(start,end +1 ):
            if i == self.page:
                self.pagination_members.append((str(i), "active", i)) 
            else:
                self.pagination_members.append((str(i), "", i)) 
        if self.page < (self.count_pages - 4):
            self.pagination_members.append(("…", "", self.page)) 

        if  self.page ==self.count_pages:
            self.pagination_members.append((str(self.count_pages), "active", self.count_pages)) 
        else:
            self.pagination_members.append((str(self.count_pages), "", self.count_pages)) 

        if self.page == self.count_pages:
            self.pagination_members.append(("next", "desabled", "+"))
        else:
            self.pagination_members.append(("next", "", "+"))
        

    def render_description(self, filters: list, description_template: str) -> str:
        """ Render the description of an object using a stored template and the current route information 

        Args:
            filters (list): List of filter in Route as WDID, where the position indicates the placeholder (e.g. filters[0] = <?f0>)
            description_template (str): the description template from templates.yml file

        Returns:
            str: the rendered description

        """        
        if not description_template:
            return ""
        
        for i,e in enumerate(filters):
            replacement: str =  '<a href="http://wikidata.org/entity/'+ e +'" target="_blank" >' + self.get_label_from_DB(e,self.language) + '</a>'
            description_template = description_template.replace("<?f" + str(i) + ">", replacement )
            description_template = description_template.replace("<?f" + str(i) + ".wdid>", e )
        return description_template  



    def render_items(self, route= ""):
        erh = ErrorHandler()
        self.route = route
        rnd = Render(route, self.response, self.language)
        r = ""
        r = '<div class ="d-flex align-content-start flex-wrap">'
        if len(self.items_on_page)>0:
            for e in self.items_on_page:
                #r += self.render_item(e)
                r += rnd.render_item(e)
        else:
            r = erh.empty(self.language)
        r += '</div>'
        self.rendered = r 


    def render_statement(self, k, v):
        prn = Presentation()
        
        if k in  ["item", "von", "bis", "image", "itemQuelle"]:
            return ""
        elif k == "itemLabel":
            rte = Route(self.route)
            if 'xml:lang' in v:
                v["route"] = rte.build_route(v['xml:lang'],  self.entity , v["wdid"])
            else:
                v["route"] = rte.build_route(self.language,  self.entity , v["wdid"])
            return prn.render_title(**v)
        elif k == "itemDescription":
            return prn.render_description(**v)
        

        elif k == "archiv":
            return prn.render_archive(**v)
        
        else:
            return prn.render_text(**v)
        
        
    def render_title(self, filters: list, title_template: str) -> str:
        """ Render the title using a stored template and the current route information 

        Args:
            filters (list): List of filter in Route as WDID, where the position indicates the placeholder (e.g. filters[0] = <?f0>)
            title_template (str): the title template from templates.yml file

        Returns:
            str: the rendered title

        """        
        if not title_template:
            return ""
        
        for i,e in enumerate(filters):
            
            title_template = title_template.replace("<?f" + str(i) + ">", self.get_label_from_DB(e,self.language ))
            title_template = title_template.replace("<?f" + str(i) + ".wdid>", e )
        return title_template            


    def render_export_link(self) -> str:
        txt: str = "Show the Wikidata query"
        query = self.query
        keywords = ("PREFIX","SELECT","WHERE","BIND","VALUES","OPTIONAL","ORDER", "SERVICE ","GROUP BY", "UNION", "FILTER", "MINUS")
        for w in keywords:
            pattern = re.compile(w, re.IGNORECASE)
            query = pattern.sub("0A"+w, query)
        query = query.replace ("0A", "\n")
        link: str = '<a href="https://query.wikidata.org/#' + quote(query) + '" class="text-white text-decoration-none" target="_blank">' + txt + '</a>'
        return link