import re, yaml, json, os
from WikidataTemplater.DBHandler import DBhandler

from WikidataTemplater.main import Settings

s = Settings()
FALLBACK_LANGUAGE= s.settings["FALLBACK_LANGUAGE"]
TEMPLATES_FILE = s.settings["TEMPLATES_FILE"]
TEMPLATES_DIR = s.settings["TEMPLATES_DIR"]

class Query():
    ''' manage the sparql query'''
    def __init__(self):
        self.template = ""
        self.templates: list = []
        self.human_readable_name = ""
        self.query = ""
        self.filters = []
        self.placeholders = []


    def build_query(self):
        '''Composes the query from template'''
        self.query = self.template.replace("<?l>", self.language)

        count_placeholder = len(self.placeholders)
        self.filters = self.filters[-1 * count_placeholder: ]
        for i,e in enumerate(self.filters):
            
            self.query = self.query.replace("<?f" + str(i)  + ">", str(e))
        

    '''
    def get_db_connection(self):
        conn = sqlite3.connect(DATABASE)
        conn.row_factory = sqlite3.Row
        return conn
    '''
    
    def is_results_cached(self,human_readable_name):
        dbh = DBhandler()
        q = 'select cached from templates where human_readable_name ="' + human_readable_name + '"'
        r = dbh.get(q)
        if r[0] == 0:
            return False
        else:
            return True
        

    def create_table(self) -> None:
        """created an emty 'templates' inside the database
        """        
        dbh: DBhandler = DBhandler()
        st_drop: str = "DROP TABLE IF EXISTS templates" 
        st_create: str =  '''
            CREATE TABLE templates (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                created TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
                human_readable_name TEXT NOT NULL,
                title TEXT,
                description TEXT,
                template TEXT NOT NULL,
                cached INTEGER

            );
                '''
        dbh.execute(st_drop)
        dbh.execute(st_create)
        
        
    def get_placeholders_in_template(self):
 
        self.placeholders = list(set(re.findall(r'\<\?f\d>', self.template)))

    def import_templates(self) -> None:
        """ A wraper used to:
            - read the yaml file
            - create the templates table
            - fill the render table with data from the yaml file
        """        
        self.read_yaml()
        self.create_table()
        self.write_db()



    def read_yaml(self) -> None:
        """ reads the yaml files in TEMPLATE_DIR and writes the data into self.templates
        """   
        for filename in os.scandir(TEMPLATES_DIR):
            if filename.is_file():
                
                d = {}
                with open(filename) as f:
                    d = yaml.load(f, Loader=yaml.FullLoader)  
                    self.templates.append(d.copy())
        


    def set_language(self, language = FALLBACK_LANGUAGE):
        self.language = language

    
    def set_filters(self, filters):
        if isinstance(filters,str):
            self.filters = [filters]
        else:
            self.filters = filters
        
    def set_template(self,template):
        self.template = str(template)
        self.get_placeholders_in_template()

    
    def set(self,template,filters, language):
        self.set_filters (filters)
        self.set_language (language)
        self.set_template(template)
        self.build_query()
        


    def write_db(self) -> None:
        """ writes data into the 'render' table
        """ 
        dbh: DBhandler = DBhandler()

        q: str = "INSERT INTO templates (human_readable_name, title, description, template, cached) VALUES (?, ?, ?, ?, ?)"
        for template in self.templates:
            
            
            if "cached" in template:
                cached = template["cached"]
            else:
                cached = 1
            dbh.insert(q , (template["human_readable_name"], json.dumps(template["title"]), json.dumps(template["description"]), template["template"],cached))