import yaml, os, sys,re, typing
import sqlite3, json, urllib.parse
from WikidataTemplater.DBHandler import DBhandler
from WikidataTemplater.main import Settings
from WikidataTemplater.DataModel import DataModel
import re

s: Settings = Settings()
RENDER_FILE = s.settings["RENDER_FILE"]
MAP_URL = s.settings["MAP_URL"]
DATA_TYPE_GEOLINK = s.settings["DATA_TYPE_GEOLINK"]


class Render():
    """renders the data into a card
       using the render.yml templates
    """    
    def __init__(self, route : str | None = None, response: dict | None = None, language: str | None = None) -> None: 
        """class initiatlization
            If calld without arguments it loads the yml templates into the database
        Args:
            route (str | None, optional): the route. Defaults to None.
            response (dict | None, optional): the Wikidata response. Defaults to None.
            language (str | None, optional): iso 639-1 two letter code. Defaults to None.
        """        
        self.templates: dict = {}
        self.template_variables: list[str] = []
        self.template_variables_keys: list[str] = []
        self.keys_not_to_show: list[str] = ["isPerson"]
        self.route: str = route
        self.response_keys: list[str] = []
        self.other_keys: list[str] = []
        self.data_model_keys: list[str] = []
        self.lines: list[str] = []
        self.original_lines: list[str] = []
        self.language: str = language
        if response:
            self.head: list[str] = response["head"].copy()
        if route:
            self.get_template(route)
            self.get_template_variables()
            self.get_template_variables_keys()
            self.get_response_keys(response)
            self.get_other_keys()
            
            
    
    
    def build_substitute (self, item: dict,  finding: str) -> str:
        """ some kind of logic usefull to display special situations.
            the operations needs to be added by dot to the value.
            Several operations may be connected, but their order has 
            to been chosen carefully.
            Currently the following operatons are implemented:
            - .split : split the value into an array. Seperator is "|"
            - .len : returns the length of a countable object
            - .last : returns the last part of a path separated by "/"

        Args:
            item (dict): the item to be displayed
            finding (str): the placeholder in render.yml wich contains 
                            the necessary information. 
                            (e.g. ?itemSource.value.split.len  – 
                            ths example will just show the number of parts
                            seperated by "|" inside the "value" statement of the
                            item["itemSource"] element)

        Returns:
            str: the result of the diffenrent operations
        """        
        element: str = item[finding[0][1:]][finding[1]]
        commands: list[str] = finding[2:]
        for e in commands:
            if e == "split":
                element = element.split("|")
            if e == "len":
                element = len(element)
            if e == "last":
                element = element.split("/")[-1:][0]
            if e == "first_is_uri":
                if len(element)>2:
                    element = '<a href="'+element[0]+'" class="text-decoration-none text-primary" target="_blank" > ' + element[1]+ '</a> <span class=" badge bg-success ms-2">' + element[2] + '</span>' 
                else:
                    element = '<a href="'+element[0]+'" class="text-decoration-none text-primary" target="_blank" > ' + element[1]+ '</a>'
            if e == "uniq":
                element = list(set(element))
            if e == "join":
                element = ", ".join(element) 
            if e == "geolink":
                if element:
                    value = element.replace("Point", "")
                    coord = re.findall (r'-?[0-9]*\.[0-9]*', element)
                    map_string = "?mlat="+ coord[1] +"&mlon="+ coord[0] +"#map=7/"+ coord[1] +"/"+ coord[0] 
                    element = '<a href= "' + MAP_URL + map_string + '" class="text-decoration-none text-primary" target="_blank" > ' 
                    element += '<i class="fa-solid fa-location-dot me-2"></i>' + value + "</a>"
        return str(element)
        
    def build_template_descriptions(self):
        """build data template description from SPARQL statement
        """        
        ############## WORK IN PROGRESS #####################    
        dbh = DBhandler()
        sql = "Select id, human_readable_name, template from templates"
        r = dbh.get(sql, "fetchall")
        for record in r:
            a = record[2].split("\r")
            for e in a:
                pattern = r"wdt:P[0-9]*|wd:Q[0-9]*|\?item|\<\?f[0-9]\>"
                x = re.findall(pattern, e, re.MULTILINE)
                
        
    
    def clean_lines(self) -> None:
        """removes all unused placeholders
        """        
        for i, line in enumerate(self.lines):
            findings: list[str] = re.findall(r'\ \?[a-zA-Z\.]*"?',line)
            original_findings: list[str] = re.findall(r'\?[a-zA-Z\.]*',line)
            if findings:
                if len(findings) != len(original_findings) :
                    for finding in findings:
                        self.lines[i] = self.lines[i].replace(finding, "")
                else:
                    self.lines[i] = ""
                
            
    
    def create_table(self) -> None:
        """created an emty 'render' inside the database
        """        
        dbh: DBhandler = DBhandler()
        st_drop: str = "DROP TABLE IF EXISTS render" 
        st_create: str =  '''
                CREATE TABLE render (
                    route TEXT NOT NULL,
                    template TEXT NOT NULL
                );
                '''
        dbh.execute(st_drop)
        dbh.execute(st_create)
    
    
 
    
    def get_label(self,item: dict, line: str , list_of_findings: list[str]) -> str:
        """ matches the a placeholder with the corresponding key within the item dict

        Args:
            item (dict): the item to be displayed
            line (str): a line of the render template
            list_of_findings (list[str]): a list of placeholders found inside the line

        Returns:
            str: the replacement of the placeholder if there is a corresponding key
        """        
        dtm = DataModel()
        for e in list_of_findings:
            a: list[str] = e.split(".")
            
            label: str = a[0][1:]
            if label in item:
                if len(a)>2:
                    substitute: str = self.build_substitute(item,a)
                else:
                    substitute: str = item[label][a[1]]
                line: str = line.replace(e,substitute)
                labels : list = re.findall(r'#[a-zA-ZääöüÄÖÜß]*',line)
                if len(labels) > 0:
                    for label in labels:
                        line = line.replace(label,dtm.get_entity_label("?"+label[1:], self.language))
                
        return line



    def get_other_keys(self) -> None:
        """ this one will find those keys within the item dict
            who have not a direct placeholder and therefore will be added
            in a new copy of the line containing the ?default.value or
            ?default.uri placeholder.
            Writes into self.other_keys 
        """        
        dtm: DataModel = DataModel()
        
        other_keys_from_data_model: list[str] = [item for item in self.response_keys if item not in self.head["vars"] and item[0] == "?"]
        other_keys_from_wd_query: list[str] = [item for item in self.response_keys if item in self.head["vars"] and item[0] != "?"]
        other_keys_from_wd_query: list[str] = [x for x in other_keys_from_wd_query if x not in self.template_variables_keys and x not in self.keys_not_to_show ]
        r: list[tuple] = dtm.select((None, "?label_" + self.language, None))
        other_keys_from_data_model_new: list[str] = []
        for e in r:
            if e[0] in other_keys_from_data_model:
                other_keys_from_data_model_new.append(e[0])
        self.other_keys = other_keys_from_wd_query 
        self.data_model_keys = other_keys_from_data_model_new
        


    def get_response_keys(self, response: dict) -> None:
        """ retrieves a set of keys which are present 
            in at least on item of the response
            Writes into self.response_keys

        Args:
            response (dict): the Wikidata response
        """
        keys: list[str] = []
        for e in response["results"]["bindings"]:
            keys.extend(e.keys())
        self.response_keys =  tuple(set(keys))


    def get_template(self, route: str) -> None:
        """ retrieves the render template from database
            depending from the given route

        Args:
            route (str): the route path for which the template should be used

        """        
        dbh: DBhandler = DBhandler()
        q: str = "select template from render where route = '" + route + "'"
        r: dict | list = dbh.get(q)
        if r:
            self.template = r[0]
        else:
            q: str = "select template from render where route = '/default'"
            r: dict | list = dbh.get(q)
            self.template = r[0]

    
    def get_template_variables(self) -> None:
        """ retrieves all placeholders used in a render template.
            Writes into self.template_variables
        """        
        self.template_variables = re.findall(r'\?[a-zA-Z\.]*',self.template)
    
    def get_template_variables_keys(self) -> None:
        """ retrieves a list of keys corresponding the the 
            placeholders found inside the temlate.
            Writes into self.temlate_varables_keys
        """        
        self.template_variables_keys = [x.split(".")[0].split("?")[1] for x in self.template_variables]
    
    def import_templates(self) -> None:
        """ A wraper used to:
            - read the yaml file
            - create the render table
            - fill the render table with data from the yaml file
        """        
        self.read_yaml()
        self.create_table()
        self.write_db()
        
    def is_uri(self,string:str) -> bool:
        
        """checks if a given string represents an uri

        Args:
            s (str): the given string

        Returns:
            bool: the result of analisys
        """        
        regex = r"(?i)\b((?:https?://|www\d{0,3}[.]|[a-z0-9.\-]+[.][a-z]{2,4}/)(?:[^\s()<>]+|\(([^\s()<>]+|(\([^\s()<>]+\)))*\))+(?:\(([^\s()<>]+|(\([^\s()<>]+\)))*\)|[^\s`!()\[\]{};:'\".,<>?«»“”‘’]))"
        url = re.findall(regex, string)
        if len(url) == 1:
            return True
        else:
            return False
            
    def read_yaml(self) -> None:
        """ reads the yaml file and writes the data into self.templates
        """        
        with open(RENDER_FILE) as f:
            self.templates= yaml.load(f, Loader=yaml.FullLoader)  

    def render_cell(self,d : dict, fieldname: str, tvw: object) -> str:
        if d['value'] is None:
            return " "
        if fieldname == "item":
            s = self.strip_wdid_from_uri(d['value'])
            href = d['value']
            img = ""
            #if d['value'][0:1] == "Q" and d['value'][1:].isnumeric():
            #if d['value'][0:1] == "Q":
            img_link = "/static/images/wikidata-logo.svg"
            img = "<img src='" + img_link + "' style='width: 2rem; '  class='me-1' / >"
            #return  img + '<a href=" ' + href  + '" target = "_blank">' + s + '</a>'
        if fieldname == "image":
            #return "<img src='" + d['value'] + "'  class=' img-fluid mx-auto rounded ' style='max-height: 120px;'/>"
            pass
        if fieldname == "uri":
            if d['value']:
                span = '<span title="'+d['value']+'">' + self.truncate(d['value']) + '</span>'
                #return '<a href=" ' + d['value']  + '" target = "_blank" >' + span + '</a>'
            else:
                pass
                #return ""
        
        if 'type' in d and d['type'] == "literal":  
            if 'datatype' in d and d['datatype'] == DATA_TYPE_GEOLINK:
                value = d['value'].replace("Point", "")
                coord = re.findall (r'-?[0-9]*\.[0-9]*', d['value'])
                map_string = "?mlat="+ coord[1] +"&mlon="+ coord[0] +"#map=7/"+ coord[1] +"/"+ coord[0] 
                element = '<a href= "' + MAP_URL + map_string + '" class="" target="_blank" > ' 
                element += '<i class="fa-solid fa-location-dot me-2"></i>' + value + "</a>"
                #return element
            elif d['value'].find("|") > -1 :
                a = d['value'].split("|")
                regex = r"(?i)\b((?:https?://|www\d{0,3}[.]|[a-z0-9.\-]+[.][a-z]{2,4}/)(?:[^\s()<>]+|\(([^\s()<>]+|(\([^\s()<>]+\)))*\))+(?:\(([^\s()<>]+|(\([^\s()<>]+\)))*\)|[^\s`!()\[\]{};:'\".,<>?«»“”‘’]))"
                url = re.findall(regex, a[0])
                if url and len(a)==2:
                    return '<a href="' +a[0]+'" target="_blank">' + a[1] + '</a>'
                else:
                    return "<ul><li>" + "</li><li>".join(a) + '</li><ul>'
            else:
                if not d['value']:
                    return ""
                else:
                    return d['value'] 
        return d['value']
        
    
    def xrender_description(self,description: str) -> str:
        """parses the description line sotred in template file and convert it to html

        Args:
            desription (str): the Description stored in the template file

        Returns:
            str: the html converted description
        """        
        a = description.split("|")
        s = ""
        if len(a) > 1:
            s += '<ul class="list-group  mb-3  " style="width: 30rem; ">'
            w = '<a href="https://www.wikidata.org/wiki/Wikidata:Tours/" target="_blank" class="text-decoration-none fw-bolder"> Wikidata Model </a>'
            s += '<li class="list-group-item list-group-item-secondary d-flex justify-content-between align-items-center" aria-current="true">'+w+'</li>'
            for e in a:
                ee = e.split("=")
                if len(ee) > 1:
                    s += '<li class="list-group-item d-flex justify-content-between align-items-center">' 
                    s += ee[0] + '<span>' + ee[1] + '</span>'
                    s += '</li>'
            s +='</ul>'
        else:
            s = a[0]
        return s
    
    def render_cache_created(self, created:str) -> str:
        """
        renders the age of cashed results below the main table
        """
        if created:
            r = "<span class='text-primary'>Data were requested from Wikidata on: </span><span class='text-secondary'> " + created + "</span>"
        else:
            r = "<span class='text-success'>The data has just been requested from Wikidata</span>"
        return r
    
    def render_description(self,description: str) -> str:
        """parses the description line to a suitable bootsrap table data object

        Args:
            desription (str): the Description stored in the template file

        Returns:
            str: the html converted description
        """     
        a = description.split("|")
        r = []
        for e in a:
            d = {}
            l = e.split("=")
            d["predicate"] =l[0].strip()
            if len(l) > 1:
                d["object"] = l[1].strip()
            r.append(d.copy())
        return json.dumps(r)
    
    def render_item(self,item: dict ) -> None:
        """ A Wrapper, used to:
           - transpone the render template linewise into self.lines
           - make a copy of self.lines
           - render keys who have a corresponding placeholder
           - render static translation using the data model
           - render keys who have not a corresponding placeholder (using the ?default)
           - remove the placeholder without corresponding keys

        Args:
            item (dict): item to b displayed

        Returns:
            str: the html string to be displayed

        """        
        self.lines: list[str] = self.template.split("\n")
        self.original_lines: list[str] = self.lines.copy()
        self.render_matches(item)
        self.render_translations(item)
        self.render_other_keys(item)
        self.render_data_model_keys(item)
        self.clean_lines()         
        return "\n".join(self.lines)
    
    def render_data_model_keys(self,item:dict):
        dtm = DataModel()
        default_lines: list[tuple[int,str]] = [(i,e) for i,e in enumerate(self.lines) if e.find("?sublink") > -1 ]
        i: int = len(default_lines) -1 
        while i > -1:
            for ii, key in enumerate(reversed(self.data_model_keys)):
                modified_line: str = self.lines[default_lines[i][0]]
                
                if key in item:
                    modified_line: str = self.lines[default_lines[i][0]]
                    modified_line = modified_line.replace("?sublink.uri", item[key]["uri"])
                    modified_line = modified_line.replace("?sublink.value", item[key]["label"])
                
                self.lines.insert( default_lines[i][0]+1,modified_line)
            i -= 1
            

    def render_matches(self, item: dict):
        """ finds placeholders within a line and calls for the label
            Modifies the self.lines list

        Args:
            item (dict): the item to be displayed
        """
        for i, line in enumerate(self.lines):
            findings: list[str]= re.findall(r'\?[a-zA-Z0-9_\.]*',line)
            if findings:
                self.lines[i] = self.get_label(item, line, findings)    
            
    def render_other_keys(self, item: dict) -> None:
        """ manages those keys who have not corresponding placeholders ad
            therefore will be displayed via the ?default placeholder

        Args:
            item (dict): the item to be displayed
        """        
        dtm = DataModel()
        default_lines: list[tuple[int,str]] = [(i,e) for i,e in enumerate(self.lines) if e.find("?default") > -1 ]
        i: int = len(default_lines) -1 
        while i > -1:
            for ii, key in enumerate(reversed(self.other_keys)):
                modified_line: str = self.lines[default_lines[i][0]]
                if key in item:
                    
                    values = item[key]["value"].split("|")
                    if len(values) == 1:
                        if self.is_uri(values[0]):
                            modified_line = modified_line.replace("?default.uri", values[0])
                            modified_line = modified_line.replace("?default.value", values[0])
                        else:
                            if "uri" in item[key]:
                                modified_line = modified_line.replace("?default.uri", item[key]["uri"])
                            else:
                                modified_line = re.sub('<\/?a[^>]*>', '', modified_line)
                            value = item[key]["value"]
                            if key in ("itemAltLabel"):
                                value = value.replace(", ", "<br />")
                            modified_line = modified_line.replace("?default.value", value)
                    else:
                        if self.is_uri(values[0]):
                            modified_line = modified_line.replace("?default.uri", values[0])
                            if len(values) > 1:
                                if len(values)>2:
                                    value = values[1] + "  ···<span class='text-success ms-2'>"+values[2]+"</span>"
                                    modified_line = modified_line.replace("?default.value", value)
                                modified_line = modified_line.replace("?default.value", values[1])
                            else:
                                modified_line = modified_line.replace("?default.value", values[0])
                        else:
                            modified_line = re.sub('<\/?a[^>]*>', '', modified_line)
                            modified_line = modified_line.replace("?default.value", "<br/>".join(values))
                modified_line = modified_line.replace("#default",dtm.get_entity_label("?"+key, self.language))
                self.lines.insert( default_lines[i][0]+1,modified_line)
            i -= 1
            
    def render_subqueries (self,item_value,route,subqueries,language,row) -> str:
        dtm = DataModel()
        a = []
        for e in subqueries:
            if "itemLabel" in row:
                item :str = "_".join(row["itemLabel"].split(" "))
            else:
                item :str = "_".join(item_value.split(" "))
            href = "/" +route+"/" + e[1:]+ ";" + item + ";" + item_value + "/" + language
            label = dtm.get_entity_label(e,language)
            #a.append ('<a href="'+href + '" onclick="javascript:spin()">' + label + '</a>')
            a.append({"uri":href, "label": label})
        #return " | ".join(a)
        return json.dumps(a)
    
    
    def render_translations(self,item: dict) -> None:
        """ render the static translation ithin the 
            render template using the data model.
            Modified the self.lines object 

        Args:
            item (dict): the item to be displayed
        """        
        dtm: DBhandler = DataModel()
        #r: list[str] = []
        for i, line in enumerate(self.lines):
            modified_line: str = line
            findings: list[str] = re.findall(r'__\([a-z0-9A-Z\ ]*\)',line)
            if findings:
                for finding in findings:
                    subject: str = "?" + re.findall(r'__\([a-z0-9A-Z\ ]*\)',finding)[0][3:-1].lower().replace(" ", "_")
                    translation: str =  dtm.get_entity_label(subject, self.language)
                    modified_line: str = modified_line.replace(finding, translation)
            self.lines[i] = modified_line

    def strip_wdid_from_uri(self, uri: str) -> str:
        """ strips the path from Wikidata identifier.
            keeps the shortcut identifier (e.g. 'Q5')

        Args:
            uri (str): Wikidata uri with path (e.g. 'http://wikidata.org/enitity/Q42') 

        Returns:
            str: shortcut Wikidata identifier (e.g. 'Q42')
        """        
        return uri.split("/")[-1:][0]

    def truncate (self, text: str, type: str = "uri") -> str:
        if type == "uri":
            return text.strip("/").split("/")[-1:][0]
        else:
            return text[0:15] + "…"

    def write_db(self) -> None:
        """ writes data into the 'render' table
        """ 
        dbh: DBhandler = DBhandler()
        q: str = "insert into render (route, template) values (?,?)"
        for k,v in self.templates.items():
            dbh.insert(q,(k,str(v)))
            
            
            
 