from WikidataTemplater.View import View
from WikidataTemplater.Presentation import Presentation
from WikidataTemplater.Route import Route
from WikidataTemplater.WDhandler import WDhandler
from WikidataTemplater.Render import Render
from WikidataTemplater.DBHandler import DBhandler
from WikidataTemplater.ErrorHandler import ErrorHandler

from urllib.parse import quote
import re


class TableView(View):
    def __init__(self, view_instance ):
        vars(self).update(vars(view_instance))
        #super.__init__()
        if len(self.subqueries) > 0: 
            self.header.append(("subqueries","Links"))
            
        self.table_data = self.get_table_data()
        self.export_link = self.render_export_link()
        
        
        
        


    def get_table_data(self):
        rnd = Render()
        self.created = rnd.render_cache_created(self.created)
        data = []
        for e in self.response["results"]["bindings"]:
            
            row = {}
            for field in self.table_header:
                
                if field[0] in e.keys():
                    row[field[0]] = e[field[0]]["value"]
            if len(self.subqueries) > 0:
                row["subqueries"] = rnd.render_subqueries (self.strip_wdid_from_uri(e['item']['value']),
                                                          self.strip_language_from_route(),
                                                          self.subqueries,
                                                          self.language, row)
            data.append(row.copy())
        return data.copy()
        
            



        
    def render_title(self, filters: list, title_template: str) -> str:
        """ Render the title using a stored template and the current route information 

        Args:
            filters (list): List of filter in Route as WDID, where the position indicates the placeholder (e.g. filters[0] = <?f0>)
            title_template (str): the title template from templates.yml file

        Returns:
            str: the rendered title

        """        
        if not title_template:
            return ""
        
        for i,e in enumerate(filters):
            span_start = '<span class="text-success">'
            span_end = "</span>"
            title_template = title_template.replace("<?f" + str(i) + ">", span_start + self.get_label_from_DB(e,self.language ) + span_end)
            title_template = title_template.replace("<?f" + str(i) + ".wdid>", span_start+ e + span_end)
        return title_template            


    def render_export_link(self) -> str:
        txt: str = "Show the Wikidata query"
        query = self.query
        keywords = ("PREFIX","SELECT","WHERE","BIND","VALUES","OPTIONAL","ORDER", "SERVICE ","GROUP BY", "UNION", "FILTER", "MINUS")
        for w in keywords:
            pattern = re.compile(w, re.IGNORECASE)
            query = pattern.sub("0A"+w, query)
        query = query.replace ("0A", "\n")
        link: str = '<a href="https://query.wikidata.org/#' + quote(query) + '" class="text-white text-decoration-none" target="_blank">' + txt + '</a>'
        return link