
from WikidataTemplater.DBHandler import DBhandler
from WikidataTemplater.WDhandler import WDhandler
from WikidataTemplater.Route import Route
from WikidataTemplater.DataModel import DataModel
from WikidataTemplater.Render import Render
from WikidataTemplater.main import Settings
from datetime import datetime, timedelta
import json, typing, sys

s: Settings = Settings()
FALLBACK_LANGUAGE = s.settings["FALLBACK_LANGUAGE"]
LANGUAGES = s.settings["LANGUAGES"]
REFRESH_PERIOD = s.settings["REFRESH_PERIOD"]
DATA_MODEL_INTRO = s.settings['DATA_MODEL_INTRO']

class View():
    """manages the selection of data to present
    """    
    def __init__(self, query: str = None , 
                 route: str = None,  
                 language: str = FALLBACK_LANGUAGE,
                 title: str = None,
                 description: str = None,
                 cached: int = 1
                 ) -> None:
        """ class initialization

        Args:
            query (str): Wikidata query string
            route (str): entire path in uri
            language (str, optional): language selected for presentation. Defaults to FALLBACK_LANGUAGE.
        """  
        
        # Query string     
        self.instances_of_items_on_pages: dict[str, str] = {}
        if query: 
            self.query: str = query
        self.wdid: str = ""
        # Date of cached response
        self.created: str = ""
        # List or SingleObject
        self.type: str = ""
        # JSON response from the Wikidata endpoint or from cache database
        self.response: dict = ""
        if language:
            self.language: str = language
        else:
            self.language: str = FALLBACK_LANGUAGE
        # Page route
        if route:
            self.route: str = route
        # Page title
        self.title = title
        
        self.description = self.get_description(description)
        self.filter = "Search"
        if cached :
            self.cached = cached
        else:
            self.cached = 1
        if route:
            self.new_wdid: list = []
            self.entity: str = self.get_entity()
            self.get_sub_queries()
            self.controller()
            self.breadcrumb = self.build_breadcrumb()
            self.search_index = []
            self.export_link = ""
            self.language_links = []
            self.build_language_links()
            self.table_header = []
            self.header = self.get_headers()
            self.data_model_intro = DATA_MODEL_INTRO[self.language]
            

            
        
    
    def build_breadcrumb(self) -> list:
        """creates a list of dict with breadcrumb elements

        Returns:
            list[dict[str,str]]: List of breadcrumb elements
        """        
        breadcrumb: list[dict[str,str]] = [{"label":"Home", "uri":"/"}]
        route: str = self.strip_language_from_route()
        a: list[str] = self.route.split("/")
        for i,e in enumerate(a):
            d = {}
            a = e.split(";")
            if len(a) == 3:
                d['label'] = a[0].replace("_", " ") + ": " + a[1].replace("_", " ")
                d['uri'] = "/"+ "/".join(self.route.split("/")[:i+1]) + "/" + self.language
                breadcrumb.append(d.copy())
        breadcrumb.append({"value":a[-1:][0], "uri": self.route})
        return breadcrumb.copy()
        

    def build_data_from_subqueries(self) -> None:
        """creates a dummy dict with elements taken from the data model
           writes into self.subqueries
        """        
        # fixme: 
        dtm: DataModel = DataModel()
        if not self.response:
            return
        for e in self.response["results"]["bindings"]:
            wdid = self.strip_wdid_from_uri(e['item']['value'])
            label_encoded = e['itemLabel']['value'].lower().replace(" ", "_")
            for ee in self.subqueries:
                e[ee] = {"value": ee, "label": dtm.get_entity_label(ee,self.language)}
                e[ee]["uri"] = "/" + self.strip_language_from_route() + "/" + ee[1:] + ";" + label_encoded + ";" + wdid + "/" + self.language

    def build_language_links(self) -> None:
        uri: str = self.strip_language_from_route()
        
        for e in LANGUAGES:
            d: dict = {"uri":"/" +uri +"/" +e, "lang": e}
            self.language_links.append(d.copy())


    def clear_cache(self) -> None:
        """ clears the 'cache' table in database
        """        
        dbh: DBhandler = DBhandler()
        q : str = "DELETE FROM cache"
        dbh.delete(q)

        
    def controller(self) -> None:
        """controls the programm flow dependant from query an subquerries
           Adds the page title
        """        
        rte: Route = Route()
        #self.title = rte.to_title(self.route)
        if self.query != "":
            if not(self.get_data_from_wd()):
                pass
                
        if self.subqueries is not None :
            self.get_data_from_subqueries()

        self.save_new_wd_objects()
        if self.cache_id is None:
            self.transpone_humans()
        self.write_response_to_cache()
        self.get_search_index()
        
    def encode_quotes(self, txt:str) -> str:
        txt = txt.replace('"', '“')
        txt = txt.replace("'","`")
        return txt

    def get_cached_results(self) -> [str| None, list, str, str]:
        """ Checks if there is a cached response in database table cache. 
            Update the response an cache_id attribute 

        Returns:
            [str| None, list, str]: Triple of cached record id, cached response and cached route
        """                
        dbh: DBhandler = DBhandler()
        q: str = "SELECT * FROM cache where route = '" + self.encode_quotes(self.route) + "'"
        
        r: list[dict] = dbh.get(q)
        if not r:
            return None, [], "", None
        else:
            
            created: datetime= datetime.fromisoformat(r["created"])
            timedelta: timedelta = int((datetime.now() -created).total_seconds())
            if int(timedelta) > int(REFRESH_PERIOD):    
                return None, [], "", ""
            else:
                
                return r["id"], json.loads(r["response"]), r["route"], r["created"]

    def get_data_from_subqueries(self) -> None:
        """Wrapper for self.build_data_from_subqueries()
        """        
        self.build_data_from_subqueries()
        


    def get_data_from_wd(self) -> None:
        """ - retrieves data from Wikidata.
            - checks if retrieved objects are instances of humans (for then transpone name)
            - writes to self.response
            - send response to cache writing
        """        
        # fixme: Error handling
        self.cache_id, self.response, self.Xtitle, self.created = self.get_cached_results()
        
        if self.cache_id is None:
            wdh: WDhandler = WDhandler()
            dtm: DataModel = DataModel()
            r: dict = wdh.get(self.query)
            if isinstance(r,dict):
                self.response = r.copy()
                return True
            else:
                # response error
                return False
        else:
            return True
            
        
    def get_description(self, description) -> str:
        """calls the render function to parse the description tag of the WD template

        Returns:
            str: _description_
        """        
        if description:
            rnd = Render()
            return rnd.render_description(description)
    
    def get_entity(self) -> str:
        """retrieves the last entity from the data model from path

        Returns:
            str: entity variable (e.g ?humans)
        """        
        path: str = self.strip_language_from_route()
        path = self.route.split("/")
        if len(path) < 2 :
            return None

        entity: str =  "?" + path[-2:-1][0].split(";")[0]
        return entity

    def get_fieldnames(self) -> list:
        """Lists all keys in self.resonse bindings

        Returns:
            list: list of keys
        """  
        l = [] 
        for e in self.response["results"]["bindings"] :
            l += e.keys()
        return set(l)
        



    def get_filter(self, filter: str) -> None :
   
        bindings = [ x for x in self.response["results"]["bindings"] if x['itemLabel']['value'].find(filter) > -1]
        self.response["results"]["bindings"] = bindings.copy()
        self.filter = filter
    
    def get_headers(self):
        dtm = DataModel()
        header = []
        table_header = []
        a = self.response["head"]["vars"]
        for e in a:
            table_header.append((e,dtm.get_entity_label(e,self.language)))
            if e[0:1]!= "_":
             header.append((e,dtm.get_entity_label(e,self.language)))
        self.table_header = table_header.copy()
        return header.copy()
    
    
    def get_instances(self, items: tuple) -> None :
        i = 0
        humans:bool = self.is_human_query()
        packet_size: int = 150
        while i < len(items) - 1:
            self.get_instances_from_wd(items[i: i + packet_size + 1],humans)
            i += 150
            
    
    def get_instances_from_wd(self, items: tuple, humans: bool = False):
        """ retrievs lowest wdid between instances (P31) 
            of elements of a tuple of Wikidata object identifiers.
            Adds results to self.instances_of_items_on_pages

        Args:
            items (tuple): tuple of Wikidata identifiers (e.g. ('Q5','Q42')) 
        """   
        if humans:
            for e in items:
                if not e in self.instances_of_items_on_pages:
                    self.instances_of_items_on_pages[e] = "Q5"
        else:
                
            q: str = '''SELECT ?item ?instance WHERE {
                    VALUES ?item{'''
            for e in items:
                q += "wd:" + e + " "
            q += '''}
                ?item wdt:P31 ?instance .
            }'''
            wdh = WDhandler()
            r = wdh.get(q)
            
            instances: dict = {}
            for e in r["results"]['bindings']:
                wdid = e["item"]["value"].split("/")[-1:][0]            
                instance = "Q" + e["instance"]["value"].split("/")[-1:][0][1:]   
                if wdid in self.instances_of_items_on_pages:
                    if int(instance[1:]) < int(self.instances_of_items_on_pages[wdid][1:]):
                        self.instances_of_items_on_pages[wdid] = instance
                else:
                    self.instances_of_items_on_pages[wdid] = instance

       
    
    def get_label_from_DB (self, wdid: str, language: str = FALLBACK_LANGUAGE) -> str:
        """_summary_

        Args:
            wdid (str): _description_
            language (str, optional): _description_. Defaults to FALLBACK_LANGUAGE.

        Returns:
            str: _description_
        """        
        dbh:DBhandler = DBhandler()
        q = "select label, language from routes where wdid = '" + wdid + "'"
        r = dbh.get(q, "fetchall")
        if r:
            l: list = [x["label"] for x in r if x['language'] == language]
            if len(l) > 0:
                label = l[0]
            else:
                label = None
            if label:
                return label
            else:
                label: str =  [x["label"] for x in r if x['language'] == FALLBACK_LANGUAGE]
                if label:
                    return label[0]
                else:
                    return r[0]["label"]
        else:
            return wdid 
    
    def get_search_index(self) -> None:
        self.search_index = [x['itemLabel']['value'] for x in self.response["results"]["bindings"]]
  
    
    def get_sub_data_from_wd(self):
        pass


    
    def get_sub_queries(self) -> None:
        """ - retrieves corresponding subqueries from data model
            - writes into self.subqueries
        """        
        dtm: DataModel = DataModel()
        if self.entity:
            
            r: list[dict] = dtm.select((self.entity, "?shows", None))
            l: list = [x[2] for x in r ]
            self.subqueries = l.copy()
        else:
            self.subqueries = None

    def is_table(self) -> bool:
        """checks if the out should be a table

        Returns:
            bool: _description_
        """        
        l: list = self.response["head"]["vars"]
        if "item" in l :
            return True
        else:
            return False

    def is_card(self) -> bool:
        """ checks if the result of a query is a list
            (it should be a list if there is a head column 'item')

        Returns:
            bool: True if it is a list, else False
        """        
        l: list = self.response["head"]["vars"]
        if "item" in l :
            return True
        else:
            return False
    
    def is_human(self, wdid: str) -> bool:
        """ checks if an item represents a human (= instance_of is Q5)

        Args:
            wdid (str): Wikidata identifier (e.g. 'Q42')

        Returns:
            bool: True if it is a human, else False
        """        
        dbh: DBhandler = DBhandler()
        q: str = 'Select instance_of from routes where wdid ="' + wdid + '"'
        r: list = dbh.get(q)
        if r:
            if r[0] == "Q5":
                return True
        return False
    
    def is_human_query(self):
        """checks if the page to present is only about humans
        """    
        route = self.strip_language_from_route()
        last_tuple= route.split("/")[-1:][0]
        if last_tuple.split(";")[0].lower() == "humans":
            return True
        else:
            return False
        
            
    
    def query_is_cached(self ) -> datetime|bool:
        """checks if a specific query exists already in the cache table



        Returns:
            datetime|None: Timestamp when the query was strored if there is one 
        """        
        dbh = DBhandler()
        #q: str = "select created from cache where query ='" + self.encode_quotes(self.query) +"' and route = '" + self.encode_quotes(self.route) + "'" 
        q: str = "select created from cache where route = '" + self.encode_quotes(self.route) + "'" 
        return dbh.get(q)
    
    def render_description(self, filters: list, description_template: str) -> str:
        """ Render the description of an object using a stored template and the current route information 

        Args:
            filters (list): List of filter in Route as WDID, where the position indicates the placeholder (e.g. filters[0] = <?f0>)
            description_template (str): the description template from templates.yml file

        Returns:
            str: the rendered description

        """        
        if not description_template:
            return ""
        
        for i,e in enumerate(filters):
            #replacement: str =  '<a href="http://wikidata.org/entity/'+ e +'" target="_blank" >' + self.get_label_from_DB(e,self.language) + '</a>'
            replacement: str = self.get_label_from_DB(e,self.language)
            description_template = description_template.replace("<?f" + str(i) + ">", replacement )
            description_template = description_template.replace("<?f" + str(i) + ".wdid>", e )
        return description_template  

    def render_title(self, filters: list, title_template: str) -> str:
        """ Render the title using a stored template and the current route information 

        Args:
            filters (list): List of filter in Route as WDID, where the position indicates the placeholder (e.g. filters[0] = <?f0>)
            title_template (str): the title template from templates.yml file

        Returns:
            str: the rendered title

        """        
        if not title_template:
            return ""
        
        for i,e in enumerate(filters):
            span_start = "<span class='text-secondary'>"
            span_end = "</span>"
            title_template = title_template.replace("<?f" + str(i) + ">", span_start + self.get_label_from_DB(e,self.language ) + span_end)
            title_template = title_template.replace("<?f" + str(i) + ".wdid>", span_start+ e + span_end)
        return title_template  

    def save_new_wd_objects(self) -> None:
        """ Checks response for items which are not in route table yet.
            Adds them if found
        """        
        if self.cached == 0:
            return 
        else:
            dbh: DBhandler = DBhandler()
            response_long_wdids: list[str] = [x["item"]["value"] for x in self.response["results"]["bindings"]]
            response_wdids: list[str] = [self.strip_wdid_from_uri(x) for x in response_long_wdids]
            response_wdids_str: str = str(tuple(response_wdids)).replace(",)", ")")
            q: str = "select wdid from routes where wdid in  "  + str(response_wdids_str)
            r: list[dict] = dbh.get(q, "fetchall")
            response_database: tuple[str] = tuple(set([x[0] for x in r]))
            if len(response_wdids) != len(response_database):
                to_ask: set = set(response_wdids).difference(set(response_database))
                self.get_instances(tuple(to_ask))
                new_wd_objects: list[dict] = []
                for e in to_ask:
                    long_wdid: str = "http://www.wikidata.org/entity/" + e
                    ee: list = [x for x in self.response["results"]["bindings"] if x['item']['value']== long_wdid]
                    if len(ee) > 0 :
                        label: str = ee[0]["itemLabel"]["value"]
                        label_encoded: str = label.lower().replace(" ","_")
                        if e in self.instances_of_items_on_pages:
                            instance: str = self.instances_of_items_on_pages[e]
                        else:
                            instance: str = ""
                        if "xml:lang" in ee[0]["itemLabel"]:
                            language: str = ee[0]["itemLabel"]["xml:lang"]
                        else:
                            language: str = ""
                        values: tuple[str] = (e, label, label_encoded, language, instance)
                        q: str = "insert into routes (wdid, label, label_encoded, language, instance_of) values (?,?,?,?,?)"                
                        dbh.insert(q,values)


    def strip_language_from_route (self) -> str:
        """strips the last part of route if it is a language indicator

        Returns:
            str: stripped route string
        """        
        a: list[str] = self.route.split("/")
        if a[-1:][0].lower() in LANGUAGES:
            striped_route: str = "/".join(a[0:-1])
            return striped_route
        else:
            return self.route
    
    def strip_wdid_from_uri(self, uri: str) -> str:
        """ strips the path from Wikidata identifier.
            keeps the shortcut identifier (e.g. 'Q5')

        Args:
            uri (str): Wikidata uri with path (e.g. 'http://wikidata.org/enitity/Q42') 

        Returns:
            str: shortcut Wikidata identifier (e.g. 'Q42')
        """        
        return uri.split("/")[-1:][0]


    def transpone_human(self, label: str) -> str:
        """ cuts the last word of a name and puts them 
            at the begin of the name followed by ','.
            cuts at the second last word if the last word is shorter then 3 characters 

        Args:
            label (str): _description_

        Returns:
            str: _description_
        """        
        a = label.strip().split(" ")
        if len(a[-1:][0]) > 1:
            a = a[-1:] + [", "] + a[0:-1]
        else:
            a = a[-2:] + [", "] + a[0:-2]
        return " ".join(a).replace(" ,", ",")

    def transpone_humans(self) -> None:
        for e in self.response["results"]["bindings"] :
            wdid: str = self.strip_wdid_from_uri(e["item"]["value"])
            if self.is_human(wdid):
                e["itemLabel"]["value"] = self.transpone_human(e["itemLabel"]["value"])
                
                

    def write_response_to_cache(self) -> None:
        """writes the Wikidata response to the database
        """
        dbh: DBhandler = DBhandler() 
        dt: str = self.query_is_cached()
        if dt:
            created: datetime= datetime.fromisoformat(dt[0])
            timedelta: timedelta = int((datetime.now() -created).total_seconds())
            if timedelta > REFRESH_PERIOD:
                set: str = "response ='"  + json.dumps(self.response).replace("'","’") + "', created = '" + str(datetime.now()).split(".")[0] + "'"
                where: str =   "query = '" + self.query + "' and route = '" + self.route + "'"
                q: str = "update cache set " + set + " where " + where
                dbh.update(q)
        else:
            q: str = 'INSERT INTO cache (query, response, route) VALUES (?,?,?)'
            values: tuple = (self.query, json.dumps(self.response),self.route)
            dbh.insert(q, values)
  
    
    def write_db(self) -> None:
        """ writes data into the 'render' table
        """ 
        dbh: DBhandler = DBhandler()
        q: str = "insert into render (route, template) values (?,?)"
        for k,v in self.templates.items():
            dbh.insert(q,(k,str(v)))