

    var map = L.map('map', 
                    {
                        minZoom:1,
                        maxZoom:13
                    }).setView([52, 12], 5);

    var Esri_WorldImagery = L.tileLayer('https://server.arcgisonline.com/ArcGIS/rest/services/World_Imagery/MapServer/tile/{z}/{y}/{x}', {
        attribution: 'Tiles &copy; Esri &mdash; Source: Esri, i-cubed, USDA, USGS, AEX, GeoEye, Getmapping, Aerogrid, IGN, IGP, UPR-EGP, and the GIS User Community'
    });
    
    var TopPlusOpen_Color = L.tileLayer('http://sgx.geodatenzentrum.de/wmts_topplus_open/tile/1.0.0/web/default/WEBMERCATOR/{z}/{y}/{x}.png', {
        maxZoom: 18,
        attribution: 'Map data: &copy; <a href="http://www.govdata.de/dl-de/by-2-0">dl-de/by-2-0</a>'
    });
    
    var OSM = L.tileLayer('https://tile.openstreetmap.org/{z}/{x}/{y}.png', {
        maxZoom: 19,
        attribution: '&copy; <a href="http://www.openstreetmap.org/copyright">OpenStreetMap</a>'
    })
    
    OSM.addTo(map);
    
    var icons = [];
    var colors = ["primary","danger","black", "sucess","info", "warning", "secondary"]
 

    $.each(colors, function(i,color){
        icons.push( L.divIcon({
            html: '<i class="bi bi-diamond-fill"></i>',
            className: 'icon text-'+color+' strong lead'
        }))
    });

    var icon_groups = []
    $.each(table_data, function(i,row){
        if (row.hasOwnProperty("_coord")){
            if (! icon_groups.includes(row["_coord"])){
                icon_groups.push(row['_coord'])
            }
        }
    });
    
    var markers = L.markerClusterGroup();
    var lat_center = 0;
    var lon_center = 0;
    var max_values = [0,0];
    var min_values = [0,0];
    var zoom = 5;
    $.each(table_data, function(i,row){
      
        if (row['coord']){
            
            popup_label = []
            for (const [key, value] of Object.entries(row)) {
                popup_label.push("<strong>" + key  + ":</strong> " + Formatter(value,"","",key));
              }
            if (row.hasOwnProperty("_coord")){
                var color = get_icon_color(row['_coord']);
            }
            else{
                var color = 0;
            }
            var coord = row["coord"].replace("Point(","")
            .replace(")","")
            .split(" ").reverse();
            coord[0] = parseFloat(coord[0]);
            coord[1] = parseFloat(coord[1]);

            if (lat_center != 0) {
                lat_center = (lat_center + coord[0] ) / 2 ;
                lon_center = (lon_center + coord[1] )/ 2;
                
                if (max_values[0] < coord[0]) { max_values[0] = coord[0]};
                if (max_values[1] < coord[1]) { max_values[1] = coord[1]} ;
                if (min_values[0] > coord[0]) { min_values[0] = coord[0]} ;
                if (min_values[1] > coord[1]) { min_values[1] = coord[1]} ;

            }
            else{
                lat_center = coord[0];
                lon_center = coord[1];
                max_values = [coord[0], coord[1]];
                min_values = [coord[0], coord[1]];
                
            }
            row['marker'] = L.marker(coord, {
                                        title : row['itemLabel'],
                                        icon : icons[color]
                                       }
                                       )
    
            row['marker'].bindPopup(popup_label.join("<br>"));
            markers.addLayer(row['marker'])
        }
    
    });

    map.addLayer( markers );


    var corner1 = L.latLng(max_values[0], max_values[1]),
    corner2 = L.latLng(min_values[0],min_values[1]),
    bounds = L.latLngBounds(corner2, corner1);
    map.fitBounds(bounds);
    

    $("#map-container").attr("class","d-none");






/*************************************** */

    function buttons () {
        return {
          btnShowMap: {
            text: 'Show object location on map',
            icon: 'bi bi-globe-europe-africa',
            event: function () {
              showMap()
            },
            attributes: {
              title: 'Show object location on map'
            }
          },
        }
      }
    
    
      function customViewFormatter (data) {
        
        var template = $('#profileTemplate').html()
        var view = ''
        $.each(data, function (i, row) {
             view += template
            .replace('%IMAGE%', row.image?Formatter(row.image,"","","image"):"")
            .replace('%VON%', row.von?row.von + " – ":"")
            .replace('%BIS%', row.bis?row.bis:"")
            .replace('%ITEMLABEL%', row.itemLabel)
            .replace('%ITEMALTLABEL%', row.itemAltLabel?row.itemAltLabel:"")
            .replace('%ITEM%', Formatter(row.item,"","","item"))
            .replace('%ITEMDESCRIPTION%', row.itemDescription?row.itemDescription:"")
            .replace('%SUBQUERIES%', row.subqueries?Formatter(row.subqueries,"","","subqueries"):"");
        })
    
        return `<div class="row mx-0">${view}</div>`
    }
    
    function Formatter(value, row, index, field) {

        
        if (field.slice(0,1)== "_") {
            return ""
        }
        if (value == undefined) {
            return "";
        }
        
        if (field == "item"){
            var img_link = "/static/images/wikidata-logo.svg"
            var img = '<img src="' + img_link + '" style="width: 2rem; ""  class="me-1" / >'
            var label = value.split("/").slice(-1,)
            var a = '<a href = "' + value + '" target="_blank" > '
            return img + a + label + "</a>"
        }
        if (field=="image"){
            return "<img src='" + value + "'  class=' img-fluid mx-auto rounded ' style='max-height: 120px;'/>"
        }
        if (field=="coord"){
            var coord = value.replace("Point", "")
                         .replace("(","")
                         .replace(")","")
                         .split(" ");
            if (coord.length > 1){
                var map_string = "?mlat="+ coord[1] +"&mlon="+ coord[0] +"#map=7/"+ coord[1] +"/"+ coord[0] ;
                a = '<a href= "' + MAP_URL + map_string + '" class="" target="_blank" > ' ; 
                var i = '<i class="fa-solid fa-location-dot me-2"></i>';
                return a + i + value.replace("Point", "")+ '</a>'
            }
        }
        if (field == "subqueries"){
            var obj = JSON.parse(value);
            var arr = [];
            var a1 = '<a href="';
            var a2 = '" onclick="javascript:spin()"> '
            for (e of obj){
                arr.push(a1 + e["uri"]+a2 + e["label"] + "</a>")
            }
            return arr.join(" | ")
        }
        if (field == "uri"  || field == "wikipedia"){
            let test = value.split("|");
            span = '<span title="'+test.slice(-1,0)+'">' + truncate(value) + '</span>';
            return '<a href=" ' + test.slice(0,1) + '" target = "_blank" >' + 
            decodeURI(span)
            .replace("_"," ")
            .replace("_", " ")
            .replace("_", " ")
            .replace("_", " ")
            .replace("_"," ") + '</a>'
        }
        let test = value.split("|");
        
        if (test.length > 1){
            
            if (test.length == 2 && isValidUrl(test[0]) ){
                
                return '<a href=" ' + test[0] + '" target = "_blank" >' + test[1] + '</a>' 
            }
            else{
                return "<ul><li>" + test.join("</li><li>") + "</li></ul>"
            }
        }
        if (value.length > 200){
            return value 
            //return value.slice(0,200) + " …"
        }
        else{
            return value
        }
      }
    
      function get_icon_color(str){
        if (icon_groups.includes(str)){
            var pos = icon_groups.indexOf(str)
            if (pos > colors.length){
                return 0
            }
            else{
                return pos
            }
        }
      }

      function isValidUrl(string) {
      try {
       
        new URL(string);    
        return true;
      } catch (err) {
        return false;
      }
    }
    
      function showMap(){
        if (lat_center != 0){
            if ($("#map-container").hasClass("d-none")){
                $("#map-container").attr("class","");
                $("#table").attr("class","table table-bordered table-hover d-none");
                $(".fixed-table-pagination").attr("style","display:none");
                $(".float-right search btn-group").attr("style","display:none");
                $("button[name = 'btnShowMap'] i ").attr("class", "bg-primary");
                $("button[name = 'btnShowMap'] i ").attr("class", "bi bi-table");
                $("button[name = 'btnShowMap'] i ").attr("title", "Show data table");
            }
            else{
                $("#map-container").attr("class","d-none");
                $("#table").attr("class","table table-bordered table-hover");
                $(".fixed-table-pagination").attr("style","");
                $(".float-right search btn-group").attr("style","");
                $("button[name = 'btnShowMap'] i ").attr("class", "bg-bi bi bi-globe-europe-africa");
                $("button[name = 'btnShowMap'] i ").attr("title", "Show object location on map");
                
            }
        }

      }
    
      function truncate(str) {
        str = str.replace(/\/$/g, '');
        var arr = str.split("/");
        if (arr.length > 1){
            return arr.slice(-1,)
        }
        else{
            return arr[0] + " …"
        }
      }
    
    


    




