
import datetime
from SPARQLWrapper import SPARQLWrapper, JSON
import re
import sqlite3
import json


FALLBACK_LANGUAGE = "de"
LANGUAGES = ("de","en","fr")
DATABASE = 'database.sqlite'
WIKIDATA_ENDPOINT = 'https://query.wikidata.org/sparql'

# Number of seconds where the query result will keep in cache database
REFRESH_PERIODE = 3600


      


class class_obj():
    '''NOT IMPELENTED manages one class from classes'''
    def __init__(self, param = ""):
        self.wdid = ""
        self.level =  0
        self.labels =  []
        self.type =""
        self.type_wdid = ""
        if param[0:1].lower() == "q" and int(param[1:]) > 0:
            self.init_by_wdid(param)


    def get_db_connection(self):
        conn = sqlite3.connect(DATABASE)
        conn.row_factory = sqlite3.Row
        return conn


    def get_wdid(self):
        pass

    def get_route_label(self, wdid, language = FALLBACK_LANGUAGE):
        pass

    def init_by_wdid(self, param):
        conn = self.get_db_connection()
        q = "SELECT * FROM classes where wdid = '" + param + "'"
        t = conn.execute(q).fetchone()
        conn.close()
        if t:
            self.id =t[0]
            self.type = t[2]
            self.type_wdid = t[3]
            self.level = t[4]
            self.wdid = t[5]
            self.labels = json.loads(t[6])


class DBhandler():
    '''Manages queries to the database '''
    def __init__(self):
        self.conn = sqlite3.connect(DATABASE)
        self.conn.row_factory = sqlite3.Row
    
    def __del__(self):
        self.conn.close()
    
    def get(self,query,fetch = 'fetchone'):
        if fetch == "fetchone":
            return self.conn.execute(query).fetchone()
        else:
            return self.conn.execute(query).fetchall()
        
    def insert(self, q, values):
        self.conn.execute(q,values)
        self.conn.commit()

class Presentation():
    ''' renders partial html'''
    def __init__(self):
        pass

    def render_image(self, **kwargs):
        r = '<img src="' + kwargs["value"] + '" class="card-image " / >'
        if "link" in kwargs.keys():
            r = '< a href = "'+ kwargs["link"] + '" target ="" >' + r + '</a>'
        return r
    
    def render_image_placeholder(self):
        #r = '''<svg class="bd-placeholder-img card-img-top" width="100%" height="180" xmlns="http://www.w3.org/2000/svg" role="img" aria-label="Placeholder" preserveAspectRatio="xMidYMid slice" focusable="false">
         #       <title>Placeholder</title>
         #       <rect width="100%" height="100%" fill="#eee"></rect></svg>'''
        return ""
    
    def render_title(self, **kwargs):
        r = '<h4 class = "card-title placeholder-glow"><figurecaption> '
        r += kwargs["value"]
        r += '</figurecaption></h3>'
        if 'route' in kwargs.keys():
            r = '<a href="' + kwargs['route']+ '" class="" >' + r + '</a>'
        return r

    def render_description(self, **kwargs):
        r = '<p class = "card-subtitle mb-2 text-body-secondary placeholder-glow"></subtitle>'
        r += kwargs["value"]
        r += '</subtitle></p>'
        return r
    
    def render_text(self, **kwargs):
        r = '<p class = "card-text placeholder-glow">'
        r += kwargs["value"]
        r += '</p>'
        return r
    
    def render_wd_link(self, **kwargs):
        r = '<badge href="' + kwargs["value"] + '" target= "_blank" class = "badge  btn-outline-success border-start border-top border-bottom border-success   ">'
        r += kwargs["value"].split("/")[-1:][0]
        r += '</badge>'
        return r

    def render_from_to(self, start, end):
        r = '<p>' + start + " – " + end + '</p>'
        return r

    def render_archive(self, language = FALLBACK_LANGUAGE, **kwargs):
        
        if len(kwargs["value"]) > 0:
            a = list(set(kwargs["value"].split("|")))
            
            r = ""
            for e in a:
                aa = e.split("~")
                if aa[-1:][0] == language :
                    if len(aa) == 3:
                        r += '<span href="'+ aa[0] +'"  target= "_blank" type="button" class="badge d-inline-block btn-secondary text-wrap m-2">' + aa[1] + '</span'
                    elif len(aa) == 4:
                        r += '<span href="'+ aa[0] +'"  target= "_blank" type="button" class="badge d-inline-block btn-secondary text-wrap m-2">' + aa[1] + \
                            '<div class="badge btn-primary text-wrap">' + aa[2] + ' </div></span>'
                else:
                    print (aa[-1:])
            return r
        return ""

    def render_sources(self, **kwargs):
        count_sources = len(kwargs["value"].split("|")) 
        if count_sources > 0:
            return '<span class="badge badge-success ps-3 position-relative">' + str(count_sources) +'</span>'
        else:
            return ""

class Query():
    ''' manage the sparql query'''
    def __init__(self):
        self.template = ""
        self.human_readable_name = ""
        self.query = ""
        self.filters = []


    def build_query(self):
        '''Composes the query from template'''
        self.query = self.template.replace("<?l>", self.language)
        for i,e in enumerate(self.filters):
            
            self.query = self.query.replace("<?f" + str(i)  + ">", str(e))
        

    
    def get_db_connection(self):
        conn = sqlite3.connect(DATABASE)
        conn.row_factory = sqlite3.Row
        return conn
    
    def set_language(self, language = FALLBACK_LANGUAGE):
        self.language = language
    
    def set_filters(self, filters):
        if isinstance(filters,str):
            self.filters = [filters]
        else:
            self.filters = filters
        
    def set_template(self,template):
        self.template = str(template)

    def set_template_by_human_readable_name(self,human_readable_name):
        self.human_readable_name = human_readable_name
        q = 'SELECT template FROM templates where type = "' + self.human_readable_name + '"'
        self.template = DBhandler.get(q)[0]
    
    def set(self,template,filters, language):
        self.set_filters (filters)
        self.set_language (language)
        self.set_template(template)
        self.build_query()


class Route():
    '''manages the route'''
    def __init__(self):
        self.language = ""
        self.filters = []
        self.template = ""
        self.query = ""

    def build_route(self, language, filters, query):
        path="/@"+ query + "/"
        q_insert = "INSERT INTO routes (wdid, label, label_encoded, language, instance_of) VALUES (?,?,?,?,?)"
        for e in filters:
            q = "SELECT label_encoded from routes where wdid ='" + e + "' and language = '" + language + "'" 
            dbh = DBhandler()
            s = dbh.get(q)
            
            if len(s) == 0:
                q = "SELECT label_encoded from routes where wdid ='" + e + "' and language = '" + FALLBACK_LANGUAGE + "'"
                s = dbh.get(q)
                
                if len(s) == 0:
                    wdh = WDhandler()
                    s = wdh.get_label(e,language) 
                    instance = str(wdh.get_instance(e))
                    v_insert = (e, s, s.lower().replace(" ", "_"), language, instance)
                    dbh.insert(q_insert, v_insert)
                    
                    if s is None:
                        s = e
                    else:
                        s = s.replace(" ", "_").replace("/","_").lower() + "|" + e.upper()
                else:
                    s = s + "@" + e
            else:
                s = s[0] + "@" + e
            path += s + "/"
        path += language  
        return path
                     

    def is_query(self,param):
        if param.isnumeric():
            q = "SELECT id from templates where id =" + str(param)
        else:
            q = "SELECT id from templates where human_readable_name = '" + str(param) + "'"
        dbh = DBhandler()
        r = dbh.get(q)
        if r:
            return r[0]
        else:
            return None

    def get_template(self, param):
        q = "SELECT template from templates where human_readable_name = '" + str(param) + "'"
        dbh = DBhandler()
        r = dbh.get(q)
        if r:
            return r[0][0]
        else:
            return None
    
    

    def parse_route(self, path):
        a = path.split("/")
        filters = []
        for e in a:
            if e[0] == "@":
                if self.is_query(e[1:].lower()):
                    self.template = self.get_template(e[1:].lower())
                    continue
            a2 = e.split("@")
            for e2 in a2:
                if e2[0:1].lower() == "q" and int(e2[1:]) > 0 :
                    filters.append(e2)
                    break
            else:
                if a2[0].lower() in LANGUAGES:
                    self.language = a2[0].lower()
                            
        if self.language == "":
            self.language = FALLBACK_LANGUAGE
        self.filters = filters
        return self.template, self.filters, self.language
    
    def to_title(self, route):
        ''' convert a route to a human redable title'''
        return route
     


class WDhandler():
    '''manages queries to Wikidata endpoint'''
    def __init__(self):
        self.sparql = SPARQLWrapper(WIKIDATA_ENDPOINT)

    def get(self, query):
        print ("QUERY TO WD\n" , query)

        self.sparql.setQuery(query)
        self.sparql.setReturnFormat(JSON)
        return self.sparql.query().convert().copy()

    def get_label(self, wdid, language):
        q = '''PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#> 
            PREFIX wd: <http://www.wikidata.org/entity/> 
            SELECT  *
            WHERE {
                    wd:'''+ wdid + ''' rdfs:label ?label .
                    FILTER (
                        langMatches( lang(?label), "''' + language + '''" ) || 
                        langMatches( lang(?label), "''' + FALLBACK_LANGUAGE + '''" )
                    )
                } 
            LIMIT 2'''
        self.sparql.setQuery(q)
        self.sparql.setReturnFormat(JSON)
        r = self.sparql.query().convert()
        r = r['results']['bindings'].copy()
        if len(r) == 0:
            return ""
        elif len(r) == 1:
            return r[0]['label']['value']
        else:
            for e in r:
                if e['label']['xml:lang'] == language.lower():
                    return e['label']['value']
                
    def get_instance(self, wdid):
        q = '''SELECT ?instance WHERE {
                wd:''' + wdid + ''' wdt:P31 ?instance .
            }'''
        print (q)
        self.sparql.setQuery(q)
        self.sparql.setReturnFormat(JSON)
        try:
            r = self.sparql.query().convert()
            r = [int(x["instance"]["value"].split("/")[-1:][0].split("Q")[1]) for x in r['results']['bindings']]
            if "Q5" in r:
                return "Q5"
            else: 
                return "Q" + str(min(r))
        except: 
            return ""
        
    def get_lowest_instance(self, a):
        if "Q5" in a:
            return "Q5"
        else: 
            return "Q" + str(min(a))







class View():
    def __init__(self, query, route, language = FALLBACK_LANGUAGE):
        # SPARQL query string
        self.query = query
        # List or SingleObject
        self.type = ""
        # JSON response from the Wikidata endpoint or from cache database
        self.response = ""
        self.language = language
        # Page route
        self.route = route
        # Page title
        self.controller()

    def build_places(self):
        dbh = DBhandler()
        q = "select * from routes where language = '"+ self.language + "'  and status = 'c'"
        r = dbh.get(q,'fetchall')
        bindings = []
        for e in r:
            item = {}
            item["item"] = {"value":"http://www.wikidata.org/entity/" + e["wdid"]}
            item["itemLabel"] = {"value": e['label'], 'xml:lang' : self.language  }
            bindings.append(item.copy()) 
        
        self.response = {"head":{},"results": {"bindings":bindings.copy()}}
        self.response["head"] = {"vars" : ['item', 'itemLabel']}
        
    
    def controller(self):
        ''' controls the origin of data'''
        rte = Route()
        self.title = rte.to_title(self.route)
        if self.route[0:1] == "_":
            self.get_data_from_static()
        else:
            self.get_data_from_wd()

    def get_data_from_static(self):
        class_label = self.route.split("/")[0][1:]
        if class_label == "places":
            self.build_places()
        elif class_label == "peoples":
            pass
        elif class_label == "times":
            pass


    def get_data_from_wd(self):
        self.cache_id, self.response, self.title = self.get_cached_results()
        if self.cache_id is None:
            wdh = WDhandler()
            r = wdh.get(self.query)
            if isinstance(r,dict):
                self.response = r.copy()
                self.write_response_to_cache()
            else:
                # response error
                pass

   
    def get_cached_results(self):
        '''
        Checks if there is a cached response. Update the response an cache_id attribute 
        '''
        dbh = DBhandler()
        q = "SELECT * FROM cache where route = '" + self.route + "'"
        r = dbh.get(q)
        if len(r) == 0:
            return None, [], ""
        elif r[0]["created"] == 0:
            pass
        else:
            return r[0]["id"], json.loads(r[0]["response"]), r[0]["route"]

    def is_list(self):
        '''we assume the a response is a list if there is a column "item"'''
        if "item" in self.response["head"]["vars"]:
            return True
        else:
            return False
        
    def write_response_to_cache(self):

        dbh = DBhandler()
        q = 'INSERT INTO cache (query, response, route) VALUES (?,?,?)'
        values = (self.query, json.dumps(self.response),self.route)
        dbh.insert(q, values)
    

class ListView(View):
    def __init__(self, view_instance):
        vars(self).update(vars(view_instance))
        #super.__init__()
        self.count_items = 0
        self.count_pages = 1
        self.pagination = True
        self.page = 1
        self.items_per_page = 30
        self.head = []
        self.order_by = 0
        self.order_desc = False
        self.items_on_page = []
        self.rendered = ""

    def paginate(self, page = None):
        if page:
            self.page = int(page)
        else:
            self.page = 1
        self.count_items = len(self.response["results"]["bindings"])
        
        self.count_pages = int(self.count_items / self.items_per_page) + 1
        
        if self.page > int(self.count_items / self.items_per_page) + 1:
            self.page = int(self.count_items / self.items_per_page) +1
        end = self.page * self.items_per_page 
        start = end - self.items_per_page 
        self.items_on_page = self.response["results"]["bindings"][start:end]
        


    def render_items(self):
        r = '<div class ="d-flex align-content-start flex-wrap">'
        for e in self.items_on_page:
            r += self.render_item(e)
        r += '</div>'
        self.rendered = r 


    def render_item(self, item):
        prn = Presentation()
        item['itemLabel']['wdid'] = item['item']['value'].split("/")[-1:][0]
        r = '<div class = "card p-3 mb-3 shadow-lg p-3 mb-5 bg-white rounded m-3"  style="width: 18rem;">'
        if "image" in item:
            r += prn.render_image(**item['image'])
        else:
            r += prn.render_image_placeholder()
        
        r += '<div class="card-body">'

        for k,v in item.items():
            r += str(self.render_statement(k,v))
        if "von" in item:
            if "bis" in item:
                r += prn.render_from_to(item['von']['value'], item['bis']['value'])
            else :
                r += prn.render_from_to(item['von']['value'], "")
        r  += '</div>'
        r  += '<div>'
        r += prn.render_wd_link(**item['item'])
        if "itemQuelle" in item:
            r += prn.render_sources(**item['itemQuelle'])
        r  += '</div>'
        
        r += '</div>'
        return r


    def render_statement(self, k, v):
        prn = Presentation()

        if k in  ["item", "von", "bis", "image", "itemQuelle"]:
            return ""
        elif k == "itemLabel":
            rte = Route()
            v["route"] = rte.build_route(v['xml:lang'], [v['wdid']], "statements_of_wd_object")
            return prn.render_title(**v)
        elif k == "itemDescription":
            return prn.render_description(**v)
        

        elif k == "archiv":
            return prn.render_archive(**v)
        
        else:
            return prn.render_text(**v)
        


        
class SingleObjectView(View):
    def __init__(self, view_instance):
        vars(self).update(vars(view_instance))
        #super.__init__()
        self.fields = []
        # WDid
        self.subject = ""

    
class Work():
    '''different mantainance task'''

    def __init__(self) -> None:
        pass

    def load_corpus(self):
        '''loads corpus data into route table'''
        wdh =  WDhandler()
        dbh = DBhandler()
        q = '''SELECT DISTINCT ?item ?itemLabel ?itemDescription where 
		{ ?item (wdt:P17|wdt:P19|wdt:P20|wdt:P27|wdt:P36|wdt:P119|wdt:P131|wdt:P159|wdt:P180|wdt:P189|wdt:P276|wdt:P279|wdt:P291|wdt:P361|
		wdt:P551|wdt:P740|wdt:P915|wdt:P840|wdt:P921|wdt:P937|
		wdt:P1001|wdt:P1071|wdt:P1269|wdt:P1376|wdt:P1416|
		wdt:P2341|wdt:P2541|wdt:P2647|wdt:P2650)/(wdt:P31*|wdt:P361*|wdt:P131*|wdt:P279*) wd:Q329618 . 
		SERVICE wikibase:label { bd:serviceParam wikibase:language "de,de,en". } } order by ?item'''
        r = wdh.get(q)
        for e in r["results"]['bindings']:
            wdid = e['item']['value'].split("/")[-1:][0]
            if wdid not in ("Q701025", "Q675321", "Q668294", "Q30607493", "Q1721466", "Q165008", "Q161062", "Q153963", "Q153665", "Q15057620"):
                label = e['itemLabel']['value']
                label_encoded = label.lower().replace(" ", "_")
                if 'xml:lang' in e['itemLabel']:
                    language = e['itemLabel']['xml:lang']
                else:
                    language = FALLBACK_LANGUAGE
                instance_of = wdh.get_instance(wdid)
                q = "INSERT INTO routes (wdid, label, label_encoded, language, instance_of, status) VALUES (?,?,?,?,?,?)"
                v = (wdid, label, label_encoded, language, instance_of, "g")
                dbh.insert(q,v)
                
